#!/usr/bin/python
# -*- coding: utf8
"""
Keywords reserved in any SQL standard

From http://www.postgresql.org/docs/9.4/static/sql-keywords-appendix.html
"""
sql_reserved_words = [
 'ABS',
 'ABSOLUTE',
 'ACTION',
 'ADD',
 'ALL',
 'ALLOCATE',
 'ALTER',
 'ANALYSE',
 'ANALYZE',
 'AND',
 'ANY',
 'ARE',
 'ARRAY',
 'ARRAY_AGG',
 'ARRAY_MAX_CARDINALITY',
 'AS',
 'ASC',
 'ASENSITIVE',
 'ASSERTION',
 'ASYMMETRIC',
 'AT',
 'ATOMIC',
 'AUTHORIZATION',
 'AVG',
 'BEGIN',
 'BEGIN_FRAME',
 'BEGIN_PARTITION',
 'BETWEEN',
 'BIGINT',
 'BINARY',
 'BIT',
 'BIT_LENGTH',
 'BLOB',
 'BOOLEAN',
 'BOTH',
 'BY',
 'CALL',
 'CALLED',
 'CARDINALITY',
 'CASCADE',
 'CASCADED',
 'CASE',
 'CAST',
 'CATALOG',
 'CEIL',
 'CEILING',
 'CHAR',
 'CHARACTER',
 'CHARACTER_LENGTH',
 'CHAR_LENGTH',
 'CHECK',
 'CLOB',
 'CLOSE',
 'COALESCE',
 'COLLATE',
 'COLLATION',
 'COLLECT',
 'COLUMN',
 'COMMIT',
 'CONDITION',
 'CONNECT',
 'CONNECTION',
 'CONSTRAINT',
 'CONSTRAINTS',
 'CONTAINS',
 'CONTINUE',
 'CONVERT',
 'CORR',
 'CORRESPONDING',
 'COUNT',
 'COVAR_POP',
 'COVAR_SAMP',
 'CREATE',
 'CROSS',
 'CUBE',
 'CUME_DIST',
 'CURRENT',
 'CURRENT_CATALOG',
 'CURRENT_DATE',
 'CURRENT_DEFAULT_TRANSFORM_GROUP',
 'CURRENT_PATH',
 'CURRENT_ROLE',
 'CURRENT_ROW',
 'CURRENT_SCHEMA',
 'CURRENT_TIME',
 'CURRENT_TIMESTAMP',
 'CURRENT_TRANSFORM_GROUP_FOR_TYPE',
 'CURRENT_USER',
 'CURSOR',
 'CYCLE',
 'DATALINK',
 'DATE',
 'DAY',
 'DEALLOCATE',
 'DEC',
 'DECIMAL',
 'DECLARE',
 'DEFAULT',
 'DEFERRABLE',
 'DEFERRED',
 'DELETE',
 'DENSE_RANK',
 'DEREF',
 'DESC',
 'DESCRIBE',
 'DESCRIPTOR',
 'DETERMINISTIC',
 'DIAGNOSTICS',
 'DISCONNECT',
 'DISTINCT',
 'DLNEWCOPY',
 'DLPREVIOUSCOPY',
 'DLURLCOMPLETE',
 'DLURLCOMPLETEONLY',
 'DLURLCOMPLETEWRITE',
 'DLURLPATH',
 'DLURLPATHONLY',
 'DLURLPATHWRITE',
 'DLURLSCHEME',
 'DLURLSERVER',
 'DLVALUE',
 'DO',
 'DOMAIN',
 'DOUBLE',
 'DROP',
 'DYNAMIC',
 'EACH',
 'ELEMENT',
 'ELSE',
 'END',
 'END-EXEC',
 'END_FRAME',
 'END_PARTITION',
 'EQUALS',
 'ESCAPE',
 'EVERY',
 'EXCEPT',
 'EXCEPTION',
 'EXEC',
 'EXECUTE',
 'EXISTS',
 'EXP',
 'EXTERNAL',
 'EXTRACT',
 'FALSE',
 'FETCH',
 'FILTER',
 'FIRST',
 'FIRST_VALUE',
 'FLOAT',
 'FLOOR',
 'FOR',
 'FOREIGN',
 'FOUND',
 'FRAME_ROW',
 'FREE',
 'FROM',
 'FULL',
 'FUNCTION',
 'FUSION',
 'GET',
 'GLOBAL',
 'GO',
 'GOTO',
 'GRANT',
 'GROUP',
 'GROUPING',
 'GROUPS',
 'HAVING',
 'HOLD',
 'HOUR',
 'IDENTITY',
 'IMMEDIATE',
 'IMPORT',
 'IN',
 'INDICATOR',
 'INITIALLY',
 'INNER',
 'INOUT',
 'INPUT',
 'INSENSITIVE',
 'INSERT',
 'INT',
 'INTEGER',
 'INTERSECT',
 'INTERSECTION',
 'INTERVAL',
 'INTO',
 'IS',
 'ISOLATION',
 'JOIN',
 'KEY',
 'LAG',
 'LANGUAGE',
 'LARGE',
 'LAST',
 'LAST_VALUE',
 'LATERAL',
 'LEAD',
 'LEADING',
 'LEFT',
 'LEVEL',
 'LIKE',
 'LIKE_REGEX',
 'LIMIT',
 'LN',
 'LOCAL',
 'LOCALTIME',
 'LOCALTIMESTAMP',
 'LOWER',
 'MATCH',
 'MAX',
 'MAX_CARDINALITY',
 'MEMBER',
 'MERGE',
 'METHOD',
 'MIN',
 'MINUTE',
 'MOD',
 'MODIFIES',
 'MODULE',
 'MONTH',
 'MULTISET',
 'NAMES',
 'NATIONAL',
 'NATURAL',
 'NCHAR',
 'NCLOB',
 'NEW',
 'NEXT',
 'NO',
 'NONE',
 'NORMALIZE',
 'NOT',
 'NTH_VALUE',
 'NTILE',
 'NULL',
 'NULLIF',
 'NUMERIC',
 'OCCURRENCES_REGEX',
 'OCTET_LENGTH',
 'OF',
 'OFFSET',
 'OLD',
 'ON',
 'ONLY',
 'OPEN',
 'OPTION',
 'OR',
 'ORDER',
 'OUT',
 'OUTER',
 'OUTPUT',
 'OVER',
 'OVERLAPS',
 'OVERLAY',
 'PAD',
 'PARAMETER',
 'PARTIAL',
 'PARTITION',
 'PERCENT',
 'PERCENTILE_CONT',
 'PERCENTILE_DISC',
 'PERCENT_RANK',
 'PERIOD',
 'PLACING',
 'PORTION',
 'POSITION',
 'POSITION_REGEX',
 'POWER',
 'PRECEDES',
 'PRECISION',
 'PREPARE',
 'PRESERVE',
 'PRIMARY',
 'PRIOR',
 'PRIVILEGES',
 'PROCEDURE',
 'PUBLIC',
 'RANGE',
 'RANK',
 'READ',
 'READS',
 'REAL',
 'RECURSIVE',
 'REF',
 'REFERENCES',
 'REFERENCING',
 'REGR_AVGX',
 'REGR_AVGY',
 'REGR_COUNT',
 'REGR_INTERCEPT',
 'REGR_R2',
 'REGR_SLOPE',
 'REGR_SXX',
 'REGR_SXY',
 'REGR_SYY',
 'RELATIVE',
 'RELEASE',
 'RESTRICT',
 'RESULT',
 'RETURN',
 'RETURNING',
 'RETURNS',
 'REVOKE',
 'RIGHT',
 'ROLLBACK',
 'ROLLUP',
 'ROW',
 'ROWS',
 'ROW_NUMBER',
 'SAVEPOINT',
 'SCHEMA',
 'SCOPE',
 'SCROLL',
 'SEARCH',
 'SECOND',
 'SECTION',
 'SELECT',
 'SENSITIVE',
 'SESSION',
 'SESSION_USER',
 'SET',
 'SIMILAR',
 'SIZE',
 'SMALLINT',
 'SOME',
 'SPACE',
 'SPECIFIC',
 'SPECIFICTYPE',
 'SQL',
 'SQLCODE',
 'SQLERROR',
 'SQLEXCEPTION',
 'SQLSTATE',
 'SQLWARNING',
 'SQRT',
 'START',
 'STATIC',
 'STDDEV_POP',
 'STDDEV_SAMP',
 'SUBMULTISET',
 'SUBSTRING',
 'SUBSTRING_REGEX',
 'SUCCEEDS',
 'SUM',
 'SYMMETRIC',
 'SYSTEM',
 'SYSTEM_TIME',
 'SYSTEM_USER',
 'TABLE',
 'TABLESAMPLE',
 'TEMPORARY',
 'THEN',
 'TIME',
 'TIMESTAMP',
 'TIMEZONE_HOUR',
 'TIMEZONE_MINUTE',
 'TO',
 'TRAILING',
 'TRANSACTION',
 'TRANSLATE',
 'TRANSLATE_REGEX',
 'TRANSLATION',
 'TREAT',
 'TRIGGER',
 'TRIM',
 'TRIM_ARRAY',
 'TRUE',
 'TRUNCATE',
 'UESCAPE',
 'UNION',
 'UNIQUE',
 'UNKNOWN',
 'UNNEST',
 'UPDATE',
 'UPPER',
 'USAGE',
 'USER',
 'USING',
 'VALUE',
 'VALUES',
 'VALUE_OF',
 'VARBINARY',
 'VARCHAR',
 'VARIADIC',
 'VARYING',
 'VAR_POP',
 'VAR_SAMP',
 'VERSIONING',
 'VIEW',
 'WHEN',
 'WHENEVER',
 'WHERE',
 'WIDTH_BUCKET',
 'WINDOW',
 'WITH',
 'WITHIN',
 'WITHOUT',
 'WORK',
 'WRITE',
 'XML',
 'XMLAGG',
 'XMLATTRIBUTES',
 'XMLBINARY',
 'XMLCAST',
 'XMLCOMMENT',
 'XMLCONCAT',
 'XMLDOCUMENT',
 'XMLELEMENT',
 'XMLEXISTS',
 'XMLFOREST',
 'XMLITERATE',
 'XMLNAMESPACES',
 'XMLPARSE',
 'XMLPI',
 'XMLQUERY',
 'XMLSERIALIZE',
 'XMLTABLE',
 'XMLTEXT',
 'XMLVALIDATE',
 'YEAR',
 'ZONE',
]




#!/usr/bin/python
# -*- coding: utf8
import logging
from collections import OrderedDict, namedtuple, defaultdict
import doctest
from hashlib import md5
import hashlib
import copy
from pprint import pprint
from ddlgenerator.reserved import sql_reserved_words
import re
try:
    import ddlgenerator.typehelpers as th
except ImportError:
    import typehelpers as th # TODO: can py2/3 split this
    
_illegal_in_column_name = re.compile(r'[^a-zA-Z0-9_$#]') 
def clean_key_name(key):
    """
    Makes ``key`` a valid and appropriate SQL column name:
    
    1. Replaces illegal characters in column names with ``_``
    
    2. Prevents name from beginning with a digit (prepends ``_``)
    
    3. Lowercases name.  If you want case-sensitive table
    or column names, you are a bad person and you should feel bad.
    """
    result = _illegal_in_column_name.sub("_", key)
    if result[0].isdigit():
        result = '_%s' % result
    if result.upper() in sql_reserved_words:
        result = '_%s' % key
    return result.lower()

def walk_and_clean(data):
    """
    Recursively walks list of dicts (which may themselves embed lists and dicts),
    transforming namedtuples to OrderedDicts and
    using ``clean_key_name(k)`` to make keys into SQL-safe column names
    
    >>> data = [{'a': 1}, [{'B': 2}, {'B': 3}], {'F': {'G': 4}}]
    >>> pprint(walk_and_clean(data))
        [OrderedDict([('a', 1)]),
         [OrderedDict([('b', 2)]), OrderedDict([('b', 3)])],
          OrderedDict([('f', OrderedDict([('g', 4)]))])]
    """
    # transform namedtuples to OrderedDicts
    if hasattr(data, '_fields'):
        data = OrderedDict((k,v) for (k,v) in zip(data._fields, data))
    # Recursively clean up child dicts and lists
    if hasattr(data, 'items'):
        for (key, val) in data.items():
            data[key] = walk_and_clean(val)
    elif isinstance(data, list) or isinstance(data, tuple) \
         or hasattr(data, '__next__') or hasattr(data, 'next'):
        data = [walk_and_clean(d) for d in data]
    
    # Clean up any keys in this dict itself
    if hasattr(data, 'items'):
        original_keys = data.keys()
        tup = ((clean_key_name(k), v) for (k, v) in data.items())
        data = OrderedDict(tup)
        if len(data) < len(original_keys):
            raise KeyError('Cleaning up %s created duplicates' %
                           original_keys)
    return data
                        
def _id_fieldname(fieldnames, table_name = ''):
    """
    Finds the field name from a dict likeliest to be its unique ID
    
    >>> _id_fieldname({'bar': True, 'id': 1}, 'foo')
    'id'
    >>> _id_fieldname({'bar': True, 'foo_id': 1, 'goo_id': 2}, 'foo')
    'foo_id'
    >>> _id_fieldname({'bar': True, 'baz': 1, 'baz_id': 3}, 'foo')
    """
    templates = ['%s_%%s' % table_name, '%s', '_%s']
    for stub in ['id', 'num', 'no', 'number']:
        for t in templates:
            if t % stub in fieldnames:
                return t % stub
            
class UniqueKey(object):
    """
    Provides unique IDs.
    
    >>> idp1 = UniqueKey('id', int, max=4)
    >>> idp1.next()
    5
    >>> idp1.next()
    6
    >>> idp2 = UniqueKey('id', str)
    >>> id2 = idp2.next()
    >>> (len(id2), type(id2))
    (32, <class 'str'>)
    """
    def __init__(self, key_name, key_type, max=0):
        self.name = key_name
        if key_type != int and not hasattr(key_type, 'lower'):
            raise NotImplementedError("Primary key field %s is %s, must be string or integer"
                                      % (key_name, key_type))
        self.type = key_type
        self.max = max
    def next(self):
        if self.type == int:
            self.max += 1
            return self.max
        else:
            return md5().hexdigest()
       
def unnest_child_dict(parent, key, parent_name=''):
    """
    If ``parent`` dictionary has a ``key`` whose ``val`` is a dict,
    unnest ``val``'s fields into ``parent`` and remove ``key``.
    
    >>> parent = {'province': 'Québec', 'capital': {'name': 'Québec City', 'pop': 491140}}
    >>> unnest_child_dict(parent, 'capital', 'provinces')
    >>> pprint(parent)
    {'capital_name': 'Québec City', 'capital_pop': 491140, 'province': 'Québec'}

    >>> parent = {'province': 'Québec', 'capital': {'id': 1, 'name': 'Québec City', 'pop': 491140}}
    >>> unnest_child_dict(parent, 'capital', 'provinces')
    >>> pprint(parent)
    {'capital_id': 1,
     'capital_name': 'Québec City',
     'capital_pop': 491140,
     'province': 'Québec'}

    >>> parent = {'province': 'Québec', 'capital': {'id': 1, 'name': 'Québec City'}}
    >>> unnest_child_dict(parent, 'capital', 'provinces')
    >>> pprint(parent)
    {'capital': 'Québec City', 'province': 'Québec'}
    
    """
    val = parent[key]
    name = "%s['%s']" % (parent_name, key)
    logging.debug("Unnesting dict %s" % name)
    id = _id_fieldname(val, parent_name)
    if id:
        logging.debug("%s is %s's ID" % (id, key))
        if len(val) <= 2:
            logging.debug('Removing ID column %s.%s' % (key, id))
            val.pop(id)
    if len(val) == 0:
        logging.debug('%s is empty, removing from %s' % (name, parent_name))
        parent.pop(key)
        return
    elif len(val) == 1:
        logging.debug('Nested one-item dict in %s, making scalar.' % name)
        parent[key] = list(val.values())[0]
        return
    else:
        logging.debug('Pushing all fields from %s up to %s' % (name, parent_name))
        new_field_names = ['%s_%s' % (key, child_key.strip('_')) for child_key in val]
        overlap = (set(new_field_names) & set(parent)) - set(id or [])
        if overlap:
            logging.error("Could not unnest child %s; %s present in %s"
                          % (name, key, ','.join(overlap), parent_name))
            return
        for (child_key, child_val) in val.items():
            new_field_name = '%s_%s' % (key, child_key.strip('_'))
            parent[new_field_name] = child_val
        parent.pop(key)

_sample_data = [{'province': 'Québec', 'capital': {'name': 'Québec City', 'pop': 491140}, 
                 'id': 1, 'province_id': 1,
                 'cities': [{'name': 'Montreal', 'pop': 1649519}, {'name': 'Laval', 'pop': 401553}]},
                {'province': 'Ontario', 'capital': {'name': 'Toronto', 'pop': 2615060}, 'province_id': 2,
                 'cities': [{'name': 'Ottawa', 'pop': 883391}, {'name': 'Missisauga', 'pop': 713443}]},
                {'province': 'New Brunswick', 'capital': {'name': 'Fredricton', 'pop': 56224}, 
                 'id': 3, 'province_id': 3,
                 'cities': [{'name': 'Saint John', 'pop': 70063}, {'name': 'Moncton', 'pop': 69074}]},
               ]

def all_values_for(data, field_name):
    return [row.get(field_name) for row in data if field_name in row]
                     
def unused_field_name(data, preferences):
    for pref in preferences:
        if not all_values_for(data, pref):
            return pref
    raise KeyError("All desired names already taken in %s" % self.name)    

class ParentTable(list):
    """
    List of ``dict``s that knows (or creates) its own primary key field. 
    
    >>> provinces = ParentTable(_sample_data, 'province', pk_name='province_id')
    >>> provinces.pk.name
    'province_id'
    >>> [p[provinces.pk.name] for p in provinces]
    [1, 2, 3]
    >>> provinces.pk.max
    3
    
    Now if province_id is unusable because it's nonunique:
    >>> data2 = copy.deepcopy(_sample_data)
    >>> for row in data2: row['province_id'] = 4
    >>> provinces2 = ParentTable(data2, 'province', pk_name='id', force_pk=True)
    >>> provinces2.pk.name
    'id'
    >>> [p[provinces2.pk.name] for p in provinces2]
    [1, 4, 3]
    
    """
    def is_in_all_rows(self, value):
        return len([1 for r in self if r.get(value)]) == len(self)
    
    def __init__(self, data, singular_name, pk_name=None, force_pk=False):
        self.name = singular_name
        super(ParentTable, self).__init__(data)
        self.pk_name = pk_name
        if force_pk or (self.pk_name and self.is_in_all_rows(self.pk_name)):
            self.assign_pk()
        else:
            self.pk = None
                          
    def suitability_as_key(self, key_name):
        """
        Returns: (result, key_type)
        ``result`` is True, False, or 'absent' or 'partial' (both still usable)
        ``key_type`` is ``int`` for integer keys or ``str`` for hash keys
        
        """
        pk_values = all_values_for(self, key_name)
        if not pk_values:
            return ('absent', int)  # could still use it
        key_type = type(th.best_coercable(pk_values))
        num_unique_values = len(set(pk_values))
        if num_unique_values < len(pk_values):
            return (False, None)     # non-unique
        if num_unique_values == len(self):
            return (True, key_type)  # perfect!
        return ('partial', key_type) # unique, but some rows need populating
       
    def use_this_pk(self, pk_name, key_type):
        if key_type == int:
            self.pk = UniqueKey(pk_name, key_type, max([0, ] + all_values_for(self, pk_name)))
        else:
            self.pk = UniqueKey(pk_name, key_type)

    def assign_pk(self):
        """
        
        """
        if not self.pk_name:
            self.pk_name = '%s_id' % self.name
            logging.warning('Primary key %s.%s not requested, but nesting demands it'
                            % (self.name, self.pk_name))
        (suitability, key_type) = self.suitability_as_key(self.pk_name)
        if not suitability:
            raise Exception('Duplicate values in %s.%s, unsuitable primary key'
                            % (self.name, self.pk_name))
        self.use_this_pk(self.pk_name, key_type)
        if suitability in ('absent', 'partial'):
            for row in self:
                if self.pk_name not in row:
                    row[self.pk_name] = self.pk.next()
   
     
def unnest_children(data, parent_name='', pk_name=None, force_pk=False):
    """
    For each ``key`` in each row of ``data`` (which must be a list of dicts),
    unnest any dict values into ``parent``, and remove list values into separate lists.
    
    Return (``data``, ``pk_name``, ``children``, ``child_fk_names``) where
    
    ``data`` 
      the transformed input list
    ``pk_name``
      field name of ``data``'s (possibly new) primary key
    ``children``
      a defaultdict(list) of data extracted from child lists
    ``child_fk_names``
      dict of the foreign key field name in each child 
   
    """
    possible_fk_names = ['%s_id' % parent_name, '_%s_id' % parent_name, 'parent_id', ]
    if pk_name:
        possible_fk_names.insert(0, '%s_%s' % (parent_name, pk_name.strip('_')))
    children = defaultdict(list)
    field_names_used_by_children = defaultdict(set)
    child_fk_names = {}
    parent = ParentTable(data, parent_name, pk_name=pk_name, force_pk=force_pk)
    for row in parent:
        for (key, val) in row.items():
            if hasattr(val, 'items'): 
                unnest_child_dict(parent=row, key=key, parent_name=parent_name)
            elif isinstance(val, list) or isinstance(val, tuple):
                # force listed items to be dicts, not scalars
                row[key] = [v if hasattr(v, 'items') else {key: v} for v in val]
        for (key, val) in row.items():
            if isinstance(val, list) or isinstance(val, tuple):
                for child in val:
                    field_names_used_by_children[key].update(set(child.keys()))
    for (child_name, names_in_use) in field_names_used_by_children.items():
        if not parent.pk:
            parent.assign_pk()
        for fk_name in possible_fk_names:
            if fk_name not in names_in_use:
                break
        else:
            raise Exception("Cannot find unused field name in %s.%s to use as foreign key"
                            % (parent_name, child_name))
        child_fk_names[child_name] = fk_name
        for row in parent:
            if child_name in row:
                for child in row[child_name]:
                    child[fk_name] = row[parent.pk.name]
                    children[child_name].append(child)
                row.pop(child_name)
    # TODO: What if rows have a mix of scalar / list / dict types?    
    return (parent, parent.pk.name if parent.pk else None, children, child_fk_names)
        
if __name__ == '__main__':
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Given data, automatically guess-generates DDL to create SQL tables.

Invoke with one table's worth of data at a time, from command line::

    $ ddlgenerator postgresql sourcedata.yaml

The ``-i`` flag generates INSERT statements as well::

    $ ddlgenerator -i postgresql sourcedata.yaml

or from Python::

    >>> menu = Table('../tests/menu.json')
    >>> ddl = menu.ddl('postgresql')
    >>> inserts = menu.inserts('postgresql')
    >>> all_sql = menu.sql('postgresql', inserts=True)

Use ``-k <keyname>`` or ``--key=<keyname>`` to set ``keyname`` as the table's
primary key.  If the field does not exist, it will be added.  If ``-k`` is not given,
no primary key will be created, *unless* it is required to set up child tables
(split out from sub-tables nested inside the original data).

You will need to hand-edit the resulting SQL to add indexes.

You can use wildcards to generate from multiple files at once::

    $ ddlgenerator postgresql "*.csv"

Remember to enclose the file path in quotes to prevent the shell
from expanding the argument (if it does, ddlgenerator will run
against each file *separately*, setting up one table for each).
"""
from collections import OrderedDict
import copy
import datetime
from decimal import Decimal
import doctest
import logging
import os.path
import pprint
import re
import textwrap
import sqlalchemy as sa
from sqlalchemy.schema import CreateTable
import dateutil.parser
import yaml
try:
    import pymongo
except ImportError:
    pymongo = None
from data_dispenser.sources import Source
try:
    import ddlgenerator.typehelpers as th
    from ddlgenerator import reshape
except ImportError:
    import typehelpers as th  # TODO: can py2/3 split this
    import reshape

logging.basicConfig(filename='ddlgenerator.log', filemode='w')
metadata = sa.MetaData()


class KeyAlreadyExists(KeyError):
    pass


dialect_names = '''drizzle firebird mssql mysql oracle postgresql
                   sqlite sybase sqlalchemy django'''.split()


def _dump(sql, *multiparams, **params):
    pass

mock_engines = {}
for engine_name in ('postgresql', 'sqlite', 'mysql', 'oracle', 'mssql'):
    mock_engines[engine_name] = sa.create_engine('%s://' % engine_name,
                                                 strategy='mock',
                                                 executor=_dump)                                                              

class Table(object):
    """
    >>> data = '''
    ... -
    ...   name: Lancelot
    ...   kg: 69.4
    ...   dob: 9 jan 461
    ... -
    ...   name: Gawain
    ...   kg: 69.4  '''
    >>> print(Table(data, "knights").ddl('postgresql').strip())
    DROP TABLE IF EXISTS knights;
    <BLANKLINE>
    CREATE TABLE knights (
    	name VARCHAR(8) NOT NULL,
    	kg DECIMAL(3, 1) NOT NULL,
    	dob TIMESTAMP WITHOUT TIME ZONE,
    	UNIQUE (name),
    	UNIQUE (dob)
    );
    """

    table_index = 0

    def _find_table_name(self, data):
        if not self.table_name:
            if pymongo and isinstance(data, pymongo.collection.Collection):
                self.table_name = data.name
            elif hasattr(data, 'lower'):  # duck-type string test
                if os.path.isfile(data):
                    (file_path, file_extension) = os.path.splitext(data)
                    self.table_name = os.path.split(file_path)[1].lower()
        self.table_name = self.table_name or \
                          'generated_table%s' % Table.table_index
        self.table_name = reshape.clean_key_name(self.table_name)
        Table.table_index += 1

    def __init__(self, data, table_name=None, default_dialect=None,
                 save_metadata_to=None, metadata_source=None,
                 varying_length_text=False, uniques=False,
                 pk_name=None, force_pk=False,
                 _parent_table=None, _fk_field_name=None, reorder=False,
                 loglevel=logging.WARN, limit=None):
        """
        Initialize a Table and load its data.

        If ``varying_length_text`` is ``True``,
        text columns will be TEXT rather than VARCHAR.
        This *improves* performance in PostgreSQL.

        If a ``metadata<timestamp>`` YAML file generated
        from a previous ddlgenerator run is
        provided, *only* ``INSERT`` statements will be produced,
        and the table structure
        determined during the previous run will be assumed.
        """
        logging.getLogger().setLevel(loglevel)
        self.varying_length_text = varying_length_text
        self.table_name = table_name
        self._find_table_name(data)
        # Send anything but Python data objects to
        # data_dispenser.sources.Source
        if hasattr(data, 'lower') or hasattr(data, 'read'):
            self.data = Source(data, limit=limit)
        else:
            try:
                self.data = iter(data)
            except TypeError:
                self.data = Source(data)
            
        if (    self.table_name.startswith('generated_table')
            and hasattr(self.data, 'table_name')):
            self.table_name = self.data.table_name
        self.table_name = self.table_name.lower()

        self.data = reshape.walk_and_clean(self.data)

        (self.data, self.pk_name, children, child_fk_names
            ) = reshape.unnest_children(data=self.data,
                                        parent_name=self.table_name,
                                        pk_name=pk_name,
                                        force_pk=force_pk)

        self.default_dialect = default_dialect
        self.comments = {}
        child_metadata_sources = {}
        if metadata_source:
            if isinstance(metadata_source, OrderedDict):
                logging.info('Column metadata passed in as OrderedDict')
                self.columns = metadata_source
            else:
                logging.info('Pulling column metadata from file %s'
                             % metadata_source)
                with open(metadata_source) as infile:
                    self.columns = yaml.load(infile.read())
            for (col_name, col) in self.columns.items():
                if isinstance(col, OrderedDict):
                    child_metadata_sources[col_name] = col
                    self.columns.pop(col_name)
                else:
                    self._fill_metadata_from_sample(col)
        else:
            self._determine_types()

        if reorder:
            ordered_columns = OrderedDict()
            if pk_name and pk_name in self.columns:
                ordered_columns[pk_name] = self.columns.pop(pk_name)
            for (c, v) in sorted(self.columns.items()):
                ordered_columns[c] = v
            self.columns = ordered_columns

        if _parent_table:
            fk = sa.ForeignKey('%s.%s' % (_parent_table.table_name,
                                          _parent_table.pk_name))
        else:
            fk = None

        self.table = sa.Table(self.table_name, metadata,
                              *[sa.Column(cname, col['satype'],
                                          fk if fk and (_fk_field_name == cname)
                                             else None,
                                          primary_key=(cname == self.pk_name),
                                          unique=(uniques and col['is_unique']),
                                          nullable=col['is_nullable'],
                                          doc=self.comments.get(cname))
                                for (cname, col) in self.columns.items()
                                if True
                                ])
        
        self.children = {child_name: Table(child_data, table_name=child_name,
                                           default_dialect=self.default_dialect,
                                           varying_length_text=varying_length_text,
                                           uniques=uniques, pk_name=pk_name,
                                           force_pk=force_pk,
                                           _parent_table=self, reorder=reorder,
                                           _fk_field_name=child_fk_names[child_name],
                                           metadata_source=child_metadata_sources.get(child_name),
                                           loglevel=loglevel)
                         for (child_name, child_data) in children.items()}

        if save_metadata_to:
            if not save_metadata_to.endswith(('.yml', 'yaml')):
                save_metadata_to += '.yaml'
            with open(save_metadata_to, 'w') as outfile:
                outfile.write(yaml.dump(self._saveable_metadata()))
            logging.info('Pass ``--save-metadata-to %s`` next time to re-use structure' %
                         save_metadata_to)

    def _saveable_metadata(self):
        result = copy.copy(self.columns)
        for v in result.values():
            v.pop('satype')  # yaml chokes on sqla classes
        for (child_name, child) in self.children.items():
            result[child_name] = child._saveable_metadata()
        return result

    def _dialect(self, dialect):
        if not dialect and not self.default_dialect:
            raise KeyError("No SQL dialect specified")
        dialect = dialect or self.default_dialect
        if dialect not in mock_engines:
            raise NotImplementedError("SQL dialect '%s' unknown" % dialect)
        return dialect

    _supports_if_exists = {k: False for k in dialect_names}
    _supports_if_exists['postgresql'] = _supports_if_exists['sqlite'] = True
    _supports_if_exists['mysql'] = _supports_if_exists['sybase'] = True
    def _dropper(self, dialect):
        template = "DROP TABLE %s %s"
        if_exists = "IF EXISTS" if self._supports_if_exists[dialect] else ""
        return template % (if_exists, self.table_name)

    _comment_wrapper = textwrap.TextWrapper(initial_indent='-- ', subsequent_indent='-- ')
    def ddl(self, dialect=None, creates=True, drops=True):
        """
        Returns SQL to define the table.
        """
        dialect = self._dialect(dialect)
        creator = CreateTable(self.table).compile(mock_engines[dialect])
        creator = "\n".join(l for l in str(creator).splitlines() if l.strip()) # remove empty lines
        comments = "\n\n".join(self._comment_wrapper.fill("in %s: %s" %
                                                        (col, self.comments[col]))
                                                        for col in self.comments)
        result = []
        if drops:
            result.append(self._dropper(dialect) + ';')
        if creates:
            result.append("%s;\n%s" % (creator, comments))
        for child in self.children.values():
            result.append(child.ddl(dialect=dialect, creates=creates,
                          drops=drops))
        return '\n\n'.join(result)

    table_backref_remover = re.compile(r',\s+table\s*\=\<.*?\>')
    capitalized_words = re.compile(r"\b[A-Z]\w+")
    sqlalchemy_setup_template = textwrap.dedent("""
        from sqlalchemy import create_engine, %s
        engine = create_engine(r'sqlite:///:memory:')
        metadata = MetaData(bind=engine)
                        
        %s = %s
        
        metadata.create_all()""" )
    def sqlalchemy(self, is_top=True):
        """Dumps Python code to set up the table's SQLAlchemy model"""
        table_def = self.table_backref_remover.sub('', self.table.__repr__())
        table_def = table_def.replace("MetaData(bind=None)", "metadata")
        table_def = table_def.replace("Column(", "\n  Column(")
        table_def = table_def.replace("schema=", "\n  schema=")
        result = [table_def, ]
        result.extend(c.sqlalchemy(is_top=False) for c in self.children)
        result = "\n".join(result)
        if is_top:
            sqla_imports = set(self.capitalized_words.findall(table_def))
            sqla_imports &= set(dir(sa))
            sqla_imports = sorted(sqla_imports)
            result = self.sqlalchemy_setup_template % (
                ", ".join(sqla_imports), self.table_name, result)
            result = textwrap.dedent(result)
        return result

    def django_models(self, metadata_source=None):
        sql = self.sql(dialect='postgresql', inserts=False, creates=True,
            drops=True, metadata_source=metadata_source)
        u = sql.split(';\n')

        try:
            import django
        except ImportError:
            print('Cannot find Django on the current path. Is it installed?')
            django = None

        if django:
            from django.conf import settings
            from django.core import management
            from django import setup

            import sqlite3
            import os

            db_filename = 'generated_db.db'

            conn = sqlite3.connect(db_filename)
            c = conn.cursor()
            for i in u:
                c.execute(i)

            if not settings.configured:
                settings.configure(
                    DEBUG='on',
                    SECRET_KEY='1234',
                    ALLOWED_HOSTS='localhost',
                    DATABASES = {'default' : {'NAME':db_filename,'ENGINE':'django.db.backends.sqlite3'}},
                    )
                django.setup()
            management.call_command('inspectdb', interactive=False)
            os.remove(db_filename)
        
    _datetime_format = {}  # TODO: test the various RDBMS for power to read the standard
    def _prep_datum(self, datum, dialect, col):
        if datum is None:
            return 'NULL'
        pytype = self.columns[col]['pytype']
        if pytype == datetime.datetime:
            datum = dateutil.parser.parse(datum)
        elif pytype == bool:
            datum = th.coerce_to_specific(datum)
        else:
            datum = pytype(str(datum))
        if isinstance(datum, datetime.datetime):
            if dialect in self._datetime_format:
                return datum.strftime(self._datetime_format[dialect])
            else:
                return "'%s'" % datum
        elif hasattr(datum, 'lower'):
            # simple SQL injection protection, sort of... ?
            return "'%s'" % datum.replace("'", "''")
        else:
            return datum

    _insert_template = "INSERT INTO {table_name} ({cols}) VALUES ({vals});"

    def inserts(self, dialect=None):        
        if dialect and dialect.startswith("sqla"):
            yield "conn = engine.connect()"
            yield "inserter = %s.insert()" % self.table_name
            for row in self.data:
                yield "conn.execute(inserter, **{row})".format(row=str(dict(row)))
            yield "conn.connection.commit()"
        else:        
            dialect = self._dialect(dialect)
            for row in self.data:
                cols = ", ".join(c for c in row)
                vals = ", ".join(str(self._prep_datum(val, dialect, key))
                                 for (key, val) in row.items())
                yield self._insert_template.format(table_name=self.table_name,
                                                   cols=cols, vals=vals)
            for child in self.children.values():
                for row in child.inserts(dialect):
                    yield row

    def sql(self, dialect=None, inserts=False, creates=True,
            drops=True, metadata_source=None):
        """
        Combined results of ``.ddl(dialect)`` and, if ``inserts==True``,
        ``.inserts(dialect)``.
        """
        result = [self.ddl(dialect, creates=creates, drops=drops)]
        if inserts:
            for row in self.inserts(dialect):
                result.append(row)
        return '\n'.join(result)

    def __str__(self):
        if self.default_dialect:
            return self.ddl()
        else:
            return self.__repr__()

    def _fill_metadata_from_sample(self, col):
        col['pytype'] = type(col['sample_datum'])
        if isinstance(col['sample_datum'], Decimal):
            col['satype'] = sa.DECIMAL(*th.precision_and_scale(col['sample_datum']))
        elif isinstance(col['sample_datum'], str):
            if self.varying_length_text:
                col['satype'] = sa.Text()
            else:
                col['satype'] = sa.Unicode(len(col['sample_datum']))
        else:
            col['satype'] = self.types2sa[type(col['sample_datum'])]
        return col

    types2sa = {datetime.datetime: sa.DateTime, int: sa.Integer,
                float: sa.Numeric, bool: sa.Boolean}

    def _determine_types(self):
        column_data = OrderedDict()
        self.columns = OrderedDict()
        self.comments = {}
        rowcount = 0
        for row in self.data:
            rowcount += 1
            keys = row.keys()
            for col_name in self.columns:
                if col_name not in keys:
                    self.columns[col_name]['is_nullable'] = True
            if not isinstance(row, OrderedDict):
                keys = sorted(keys)
            for k in keys:
                v = row[k]
                if not th.is_scalar(v):
                    v = str(v)
                    self.comments[k] = 'nested values! example:\n%s' % \
                                       pprint.pformat(v)
                    logging.warning('in %s: %s' % (k, self.comments[k]))
                v = th.coerce_to_specific(v)
                if k not in self.columns:
                    self.columns[k] = {'sample_datum': v,
                                       'is_nullable': not (rowcount == 1 and
                                                           v is not None),
                                       'is_unique': set([v, ])}
                else:
                    col = self.columns[k]
                    old_sample_datum = col.get('sample_datum')
                    col['sample_datum'] = th.best_representative(
                        col['sample_datum'], v)
                    if (v is None):
                        col['is_nullable'] = True
                    if (col['is_unique'] != False):
                        if v in col['is_unique']:
                            col['is_unique'] = False
                        else:
                            col['is_unique'].add(v)
        for col_name in self.columns:
            col = self.columns[col_name]
            self._fill_metadata_from_sample(col)
            col['is_unique'] = bool(col['is_unique'])


if __name__ == '__main__':
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



import argparse
import logging
try:
    from ddlgenerator.ddlgenerator import Table, dialect_names
except ImportError:
    from ddlgenerator import Table, dialect_names  # TODO: can py2/3 split this
# If anyone can explain these import differences to me, I will buy you a cookie.

def read_args():
    parser = argparse.ArgumentParser(description='Generate DDL based on data')
    parser.add_argument('dialect', help='SQL dialect to output', type=str.lower)
    parser.add_argument('datafile', help='Path to file storing data (accepts .yaml, .json)', nargs='+')
    parser.add_argument('-k', '--key', help='If primary key needed, name it this', type=str.lower)
    parser.add_argument('--force-key', help='Force every table to have a primary key',
                        action='store_true')
    parser.add_argument('-r', '--reorder', help='Reorder fields alphabetically, ``key`` first',
                        action='store_true')
    parser.add_argument('-u', '--uniques', action='store_true',
                        help='Include UNIQUE constraints where data is unique')
    parser.add_argument('-t', '--text', action='store_true',
                        help='Use variable-length TEXT columns instead of VARCHAR')

    parser.add_argument('-d', '--drops', action='store_true', help='Include DROP TABLE statements')
    parser.add_argument('-i', '--inserts', action='store_true', help='Include INSERT statements')
    parser.add_argument('--no-creates', action='store_true', help='Do not include CREATE TABLE statements')
    parser.add_argument('--limit', type=int, default=None, help='Max number of rows to read from each source file')
    parser.add_argument('--save-metadata-to', type=str, metavar='FILENAME',
                        help='Save table definition in FILENAME for later --use-saved-metadata run')
    parser.add_argument('--use-metadata-from', type=str, metavar='FILENAME',
                        help='Use metadata saved in FROM for table definition, do not re-analyze table structure')

    parser.add_argument('-l', '--log', type=str.upper,
                        help='log level (CRITICAL, FATAL, ERROR, DEBUG, INFO, WARN)', default='WARN')
    args = parser.parse_args()

    return args

def set_logging(args):
    try:
        loglevel = int(getattr(logging, args.log))
    except (AttributeError, TypeError) as e:
        raise NotImplementedError('log level "%s" not one of CRITICAL, FATAL, ERROR, DEBUG, INFO, WARN' %
                                  args.log)
    logging.getLogger().setLevel(loglevel)

def generate():
    args = read_args()
    set_logging(args)
    logging.info(str(args))
    if args.dialect in ('pg', 'pgsql', 'postgres'):
        args.dialect = 'postgresql'
    if args.dialect in ('dj', 'djan'):
        args.dialect = 'django'

    if args.dialect not in dialect_names:
        raise NotImplementedError('First arg must be one of: %s' % ", ".join(dialect_names))
    for datafile in args.datafile:
        table = Table(datafile, varying_length_text=args.text, uniques=args.uniques,
                      pk_name = args.key, force_pk=args.force_key, reorder=args.reorder,
                      save_metadata_to=args.save_metadata_to, metadata_source=args.use_metadata_from,
                      loglevel=args.log, limit=args.limit)
        if args.dialect.startswith('sqla'):
            print(table.sqlalchemy())
            if args.inserts:
                print("\n".join(table.inserts(dialect=args.dialect)))
                #inserter.compile().bindtemplate
        elif args.dialect.startswith('dj'):
            table.django_models()
        else:
            print(table.sql(dialect=args.dialect, inserts=args.inserts,
                            creates=(not args.no_creates), drops=args.drops,
                            metadata_source=args.use_metadata_from))




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Various functions for examining data types.
"""
import datetime
from decimal import Decimal, InvalidOperation
import doctest
import math
import re
import sqlalchemy as sa
import dateutil.parser

def is_scalar(x):
    return hasattr(x, 'lower') or not hasattr(x, '__iter__')

def precision_and_scale(x):
    """
    From a float, decide what precision and scale are needed to represent it.

    >>> precision_and_scale(54.2)
    (3, 1)
    >>> precision_and_scale(9)
    (1, 0)

    Thanks to Mark Ransom,
    http://stackoverflow.com/questions/3018758/determine-precision-and-scale-of-particular-number-in-python
    """
    if isinstance(x, Decimal):
        precision = len(x.as_tuple().digits)
        scale = -1 * x.as_tuple().exponent
        return (precision, scale)
    max_digits = 14
    int_part = int(abs(x))
    magnitude = 1 if int_part == 0 else int(math.log10(int_part)) + 1
    if magnitude >= max_digits:
        return (magnitude, 0)
    frac_part = abs(x) - int_part
    multiplier = 10 ** (max_digits - magnitude)
    frac_digits = multiplier + int(multiplier * frac_part + 0.5)
    while frac_digits % 10 == 0:
        frac_digits /= 10
    scale = int(math.log10(frac_digits))
    return (magnitude + scale, scale)

_complex_enough_to_be_date = re.compile(r"[\-\. /]")
def coerce_to_specific(datum):
    """
    Coerces datum to the most specific data type possible
    Order of preference: datetime, boolean, integer, decimal, float, string

    >>> coerce_to_specific(7.2)
    Decimal('7.2')
    >>> coerce_to_specific("Jan 17 2012")
    datetime.datetime(2012, 1, 17, 0, 0)
    >>> coerce_to_specific("something else")
    'something else'
    """
    if datum is None:
        return None 
    try:
        if len(_complex_enough_to_be_date.findall(datum)) > 1:
            return dateutil.parser.parse(datum)
    except Exception as e:
        pass
    if str(datum).strip().lower() in ('0', 'false', 'f', 'n', 'no'):
        return False
    elif str(datum).strip().lower() in ('1', 'true', 't', 'y', 'yes'):
        return True
    try:
        return int(str(datum))
    except ValueError:
        pass
    try:
        return Decimal(str(datum))
    except InvalidOperation:
        pass
    try:
        return float(str(datum))
    except ValueError:
        pass
    return str(datum)

def _places_b4_and_after_decimal(d):
    """
    >>> _places_b4_and_after_decimal(Decimal('54.212'))
    (2, 3)
    """
    tup = d.as_tuple()
    return (len(tup.digits) + tup.exponent, max(-1*tup.exponent, 0))

def worst_decimal(d1, d2):
    """
    Given two Decimals, return a 9-filled decimal representing both enough > 0 digits
    and enough < 0 digits (scale) to accomodate numbers like either.

    >>> worst_decimal(Decimal('762.1'), Decimal('-1.983'))
    Decimal('999.999')
    """
    (d1b4, d1after) = _places_b4_and_after_decimal(d1)
    (d2b4, d2after) = _places_b4_and_after_decimal(d2)
    return Decimal('9' * max(d1b4, d2b4) + '.' + '9' * max(d1after, d2after))

def set_worst(old_worst, new_worst):
    """
    Pad new_worst with zeroes to prevent it being shorter than old_worst.
    
    >>> set_worst(311920, '48-49')
    '48-490'
    """
    
    new_len = len(str(new_worst))
    old_len = len(str(old_worst))
    if new_len < old_len:
        new_type = type(new_worst)
        new_worst = str(new_worst).ljust(old_len, '0')
        return new_type(new_worst)
    return new_worst
    
def best_representative(d1, d2):
    """
    Given two objects each coerced to the most specific type possible, return the one
    of the least restrictive type.

    >>> best_representative(None, Decimal('6.1'))
    Decimal('6.1')
    >>> best_representative(311920, '48-49')
    '48-490'
    >>> best_representative(6, 'foo')
    'foo'
    >>> best_representative(Decimal('4.95'), Decimal('6.1'))
    Decimal('9.99')
    """
  
    if d1 is None:
        return d2
    elif d2 is None:
        return d1
    preference = (datetime.datetime, bool, int, Decimal, float, str)
    worst_pref = 0
    worst = ''
    for coerced in (d1, d2):
        pref = preference.index(type(coerced))
        if pref > worst_pref:
            worst_pref = pref
            worst = set_worst(worst, coerced)
        elif pref == worst_pref:
            if isinstance(coerced, Decimal):
                worst = set_worst(worst, worst_decimal(coerced, worst))
            elif isinstance(coerced, float):
                worst = set_worst(worst, max(coerced, worst))
            else:  # int, str
                if len(str(coerced)) > len(str(worst)):
                    worst = set_worst(worst, coerced)
    return worst

def best_coercable(data):
    """
    Given an iterable of scalar data, returns the datum representing the most specific
    data type the list overall can be coerced into, preferring datetimes, then bools,
    then integers, then decimals, then floats, then strings.

    >>> best_coercable((6, '2', 9))
    6
    >>> best_coercable((Decimal('6.1'), 2, 9))
    Decimal('6.1')
    >>> best_coercable(('2014 jun 7', '2011 may 2'))
    datetime.datetime(2014, 6, 7, 0, 0)
    >>> best_coercable((7, 21.4, 'ruining everything'))
    'ruining everything'
    """
    preference = (datetime.datetime, bool, int, Decimal, float, str)
    worst_pref = 0
    worst = ''
    for datum in data:
        coerced = coerce_to_specific(datum)
        pref = preference.index(type(coerced))
        if pref > worst_pref:
            worst_pref = pref
            worst = coerced
        elif pref == worst_pref:
            if isinstance(coerced, Decimal):
                worst = worst_decimal(coerced, worst)
            elif isinstance(coerced, float):
                worst = max(coerced, worst)
            else:  # int, str
                if len(str(coerced)) > len(str(worst)):
                    worst = coerced
    return worst

def sqla_datatype_for(datum):
    """
    Given a scalar Python value, picks an appropriate SQLAlchemy data type.

    >>> sqla_datatype_for(7.2)
    DECIMAL(precision=2, scale=1)
    >>> sqla_datatype_for("Jan 17 2012")
    <class 'sqlalchemy.sql.sqltypes.DATETIME'>
    >>> sqla_datatype_for("something else")
    Unicode(length=14)
    """
    try:
        if len(_complex_enough_to_be_date.findall(datum)) > 1:
            dateutil.parser.parse(datum)
            return sa.DATETIME
    except (TypeError, ValueError):
        pass
    try:
        (prec, scale) = precision_and_scale(datum)
        return sa.DECIMAL(prec, scale)
    except TypeError:
        return sa.Unicode(len(datum))

if __name__ == '__main__':
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

setup(
    name='ddlgenerator',
    version='0.1.7.2',
    description='Generates SQL DDL that will accept Python data',
    long_description=readme + '\n\n' + history,
    author='Catherine Devlin',
    author_email='catherine.devlin@gmail.com',
    url='https://github.com/catherinedevlin/ddl-generator',
    packages=[
        'ddlgenerator',
    ],
    package_dir={'ddlgenerator': 'ddlgenerator'},
    include_package_data=True,
    install_requires=[
      "python-dateutil",
      "sqlalchemy",
      "dateutils",
      "pyyaml",
      "data_dispenser>=0.2.4",
    ],
    license="MIT",
    zip_safe=False,
    keywords='ddlgenerator',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
    ],
    test_suite='tests',
    entry_points={
        'console_scripts': [
            'ddlgenerator = ddlgenerator.console:generate',
        ]
    }    
)
